from manim import (
    AnimationGroup,
    ApplyMethod,
    Transform,
    FadeIn,
    FadeOut
)
from manim.constants import *
import numpy as np

from ..algebra.system_of_equations import SystemOfEquations

__all__ = [
    "TransformSystem",
    "SwapEquations"
]


class TransformSystem(AnimationGroup):
    """
    Transform a system of equations to another system of equations.

    Parameters
    ----------
    system_1 : :class:`SystemOfEquations`
        The first system of equations.
    system_2 : :class:`SystemOfEquations`
        The second system of equations.

    """
    def __init__(self, system_1, system_2, **kwargs):
        self.system_1 = system_1
        self.system_2 = system_2
        group = [Transform(system_1.brace, system_2.brace)]

        equations_1 = system_1.equations
        equations_2 = system_2.equations

        size_1 = len(equations_1)
        size_2 = len(equations_2)

        if size_1 <= size_2:
            for i in range(size_1):
                group.append(Transform(equations_1[i], equations_2[i]))
            for i in range(size_1, size_2):
                group.append(FadeIn(equations_2[i]))
        else:
            for i in range(size_2):
                group.append(Transform(equations_1[i], equations_2[i]))
            for i in range(size_2, size_1):
                group.append(FadeOut(equations_1[i]))

        super().__init__(*group, **kwargs)

    def clean_up_from_scene(self, scene):
        super().clean_up_from_scene(scene)
        for eq in self.system_1.equations:
            scene.remove(eq)
        scene.remove(self.system_1.brace)
        for eq in self.system_2.equations:
            scene.remove(eq)
        scene.remove(self.system_2.brace)
        scene.add(self.system_2)


class SwapEquations(AnimationGroup):
    """
    Swap two equations in a system of equations.

    Parameters
    ----------
    system : :class:`SystemOfEquations`
        The system of equations.
    i : int
        The index of the first equation.
    j : int
        The index of the second equation.
    """
    def __init__(self, system, i, j, **kwargs):
        self.system = system
        self.i, self.j = min(i, j), max(i, j)
        equations_between = self.system.equations[self.i + 1:self.j]
        delta = self.system.equations[self.i].get_height() - self.system.equations[self.j].get_height()
        super().__init__(
            ApplyMethod(equations_between.shift, delta * UP),
            ApplyMethod(self.system.equations[self.i].shift,
                        (2 * self.system.buff_between_equations + equations_between.get_height() +
                         self.system.equations[self.j].get_height()) * DOWN),
            ApplyMethod(self.system.equations[self.j].shift,
                        (2 * self.system.buff_between_equations + equations_between.get_height() +
                         self.system.equations[self.i].get_height()) * UP),
            **kwargs
        )

    def finish(self):
        self.system.equations[self.i], self.system.equations[self.j] = (
            self.system.equations[self.j], self.system.equations[self.i])


class InsertEquation(AnimationGroup):
    """
    Insert an equation to the system of equations.

    Parameters
    ----------
    system : :class:`SystemOfEquations`
        The system of equations.
    eq : :class:`manim.mobject.Mobject`
        The equation to insert.
    index : int
        The index to insert.
    direction : np.ndarray, optional
    """
    def __init__(self,
                 system: SystemOfEquations,
                 eq,
                 index,
                 direction=DOWN
                 ):
        self.system = system
        self.eq = eq
        self.index = index
        self.direction = direction

        if np.array_equal(direction, DOWN):
            under = system.equations[index:]
            delta = system.buff_between_equations + eq.get_height()
            eq.move_to(
                system.equations[index].get_center() + (
                    system.equations[index].get_height() - eq.get_height()) / 2 * UP +
                (system.equations[index].get_width() - eq.get_width()) / 2 * LEFT
            )

            brace_height = system.brace.get_height() + delta
            brace = system.brace.copy().scale_to_fit_height(brace_height).shift(delta / 2 * DOWN)

            shift_animation = ApplyMethod(under.shift, delta * DOWN)
            transform_brace = Transform(system.brace, brace)
            fade_in_animation = FadeIn(eq)

            super().__init__(
                shift_animation,
                transform_brace,
                fade_in_animation
            )
        elif np.array_equal(direction, UP):
            over = system.equations[:index]
            delta = system.buff_between_equations + eq.get_height()
            eq.move_to(
                system.equations[index].get_center() +
                (system.equations[index].get_height() + eq.get_height()) / 2 * UP +
                + self.system.buff_between_equations * UP +
                (system.equations[index].get_width() - eq.get_width()) / 2 * LEFT
            )

            brace_height = system.brace.get_height() + delta
            brace = system.brace.copy().scale_to_fit_height(brace_height).shift(delta / 2 * UP)

            shift_animation = ApplyMethod(over.shift, delta * UP)
            transform_brace = Transform(system.brace, brace)
            fade_in_animation = FadeIn(eq)

            super().__init__(
                shift_animation,
                transform_brace,
                fade_in_animation
            )
        else:
            under = system.equations[index:]
            over = system.equations[:index]
            delta = (system.buff_between_equations + eq.get_height()) / 2

            shift_under = ApplyMethod(under.shift, delta * DOWN)
            shift_over = ApplyMethod(over.shift, delta * UP)
            eq.move_to(
                system.equations[index].get_center() + (delta * DOWN) +
                (system.equations[index].get_height() + eq.get_height()) / 2 * UP +
                + self.system.buff_between_equations * UP +
                (system.equations[index].get_width() - eq.get_width()) / 2 * LEFT
            )

            brace_height = system.brace.get_height() + delta * 2
            brace = system.brace.copy().scale_to_fit_height(brace_height)

            super().__init__(
                shift_under,
                shift_over,
                Transform(system.brace, brace),
                FadeIn(eq)
            )

    def finish(self):
        self.system.equations.insert(self.index, self.eq)



from .default import *  # noqa: F401, F403
from .cevians_and_perpendiculars import *  # noqa: F401, F403
from .intersection import *  # noqa: F401, F403
from .triangle_centers import *  # noqa: F401, F403
from .circles import *  # noqa: F401, F403



import svgelements as se

from manim import (
    VMobjectFromSVGPath,
    Mobject
)
from manim.constants import *
from manim.typing import Point3DLike

from ..useful_in_development import *
from ..geometry.default import Line

__all__ = [
    "Bracket",
    "BracketBetweenPoints"
]


class Bracket(VMobjectFromSVGPath):
    """Analogue of manim's Brace, but for a rectangle.

    Examples
    --------
    .. manimextra:: BracketExample
        :save_last_frame:

        class BracketExample(Scene):
            def construct(self):
                sq = Square(color=BLUE, fill_opacity=0.5)
                bracket = Bracket(sq, buff=0.1)
                self.add(sq, bracket)
    """
    def __init__(
            self,
            mobject: Mobject,
            direction=DOWN,
            buff=0.2,
            **kwargs
    ):
        path_string = "M0.95 0.33H0.7337V2.31H0.95"
        path = se.Path(path_string)

        self.buff = buff

        super().__init__(path_obj=path, **kwargs)

        angle = -np.arctan2(*direction[:2]) + np.pi
        mobject.rotate(-angle)

        left = mobject.get_corner(DOWN + LEFT)
        right = mobject.get_corner(DOWN + RIGHT)
        self.width = right[0] - left[0]
        self.scale_to_fit_height(self.width)

        self.rotate(PI / 2)
        self.shift(left - self.get_corner(UP + LEFT) + self.buff * DOWN)
        self.rotate(angle, about_point=mobject.get_center())


class BracketBetweenPoints(Bracket):
    """Class that creates a bracket between two points.

    Parameters
    ----------
    point_1 : Point3DLike, optional
        The first point, by default ORIGIN
    point_2 : Point3DLike, optional
        The second point, by default ORIGIN
    direction : Point3DLike, optional
        The direction of the bracket, by default ORIGIN

    Examples
    --------
    .. manimextra:: BracketBetweenPointsExample
        :save_last_frame:

        class BracketBetweenPointsExample(Scene):
            def construct(self):
                A = Dot(LEFT + DOWN)
                B = Dot(RIGHT + UP)
                bracket = BracketBetweenPoints(A, B, buff=0.1)
                self.add(A, B, bracket)

    """
    def __init__(
        self,
        point_1: Point3DLike = LEFT,
        point_2: Point3DLike = RIGHT,
        direction: Point3DLike | None = ORIGIN,
        **kwargs,
    ):
        point_1, point_2 = dot_to_array(point_1, point_2)
        if all(direction == ORIGIN):
            line_vector = point_2 - point_1
            direction = np.array([line_vector[1], -line_vector[0], 0])
            direction = direction / np.linalg.norm(direction)
        super().__init__(Line(point_1, point_2), direction=direction, **kwargs)



from .default import *


__all__ = [
    "Incenter",
    "Excenter",
    "Centroid",
    "Circumcenter",
    "Orthocenter",
    "NinePointCenter",
    "LemoinePoint",
    "GergonnePoint",
    "NagelPoint",
    "Mittenpunkt",
    "SpiekerCenter",
    "FeuerbachPoint",
    "FermatPoint"
]


def barycentric_to_cartesian(A, B, C, coordinates) -> np.ndarray:
    A, B, C, = dot_to_array(A, B, C)
    coordinates = np.array(coordinates) / np.array([sum(coordinates)] * 3)
    x = coordinates[0] * A[0] + coordinates[1] * B[0] + coordinates[2] * C[0]
    y = coordinates[0] * A[1] + coordinates[1] * B[1] + coordinates[2] * C[1]
    return np.array([x, y, 0])


def trilinear_to_cartesian(A, B, C, coordinates) -> np.ndarray:
    A, B, C, = dot_to_array(A, B, C)
    a, b, c = distance(B, C), distance(A, C), distance(B, A)
    coordinates = np.array(coordinates) * np.array([a, b, c])
    return barycentric_to_cartesian(A, B, C, coordinates)


def sec(alpha) -> float:
    return 1 / np.cos(alpha)


def csc(alpha) -> float:
    return 1 / np.sin(alpha)


class Incenter(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        super().__init__(trilinear_to_cartesian(A, B, C, (1, 1, 1)), **kwargs)


class Excenter(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        super().__init__(trilinear_to_cartesian(A, B, C, (1, -1, 1)), **kwargs)


class Centroid(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        a, b, c = distance(B, C), distance(A, C), distance(A, B)
        super().__init__(trilinear_to_cartesian(A, B, C, (1 / a, 1 / b, 1 / c)), **kwargs)


class Circumcenter(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        angle_a = Angle.from_three_points(B, A, C).get_value()
        angle_b = Angle.from_three_points(A, B, C).get_value()
        angle_c = Angle.from_three_points(B, C, A).get_value()
        super().__init__(trilinear_to_cartesian(A, B, C, (np.cos(angle_a), np.cos(angle_b), np.cos(angle_c))), **kwargs)


class Orthocenter(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        angle_a = Angle.from_three_points(B, A, C).get_value()
        angle_b = Angle.from_three_points(A, B, C).get_value()
        angle_c = Angle.from_three_points(B, C, A).get_value()
        super().__init__(trilinear_to_cartesian(A, B, C, (sec(angle_a), sec(angle_b), sec(angle_c))), **kwargs)


class NinePointCenter(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        angle_a = Angle.from_three_points(B, A, C).get_value()
        angle_b = Angle.from_three_points(A, B, C).get_value()
        angle_c = Angle.from_three_points(B, C, A).get_value()
        super().__init__(trilinear_to_cartesian(
            A, B, C, (np.cos(angle_b - angle_c), np.cos(angle_c - angle_a), np.cos(angle_a - angle_b))), **kwargs)


class LemoinePoint(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        a, b, c = distance(B, C), distance(A, C), distance(A, B)
        super().__init__(trilinear_to_cartesian(A, B, C, (a, b, c)), **kwargs)


class GergonnePoint(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        a, b, c = distance(B, C), distance(A, C), distance(A, B)
        super().__init__(trilinear_to_cartesian(A, B, C, (
            (b * c / (b + c - a)), (c * a / (c + a - b)), (a * b / (a + b - c))
        )), **kwargs)


class NagelPoint(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        a, b, c = distance(B, C), distance(A, C), distance(A, B)
        super().__init__(trilinear_to_cartesian(A, B, C, (
            ((b + c - a) / a), ((c + a - b) / b), ((a + b - c) / c)
        )), **kwargs)


class Mittenpunkt(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        a, b, c = distance(B, C), distance(A, C), distance(A, B)
        super().__init__(trilinear_to_cartesian(A, B, C, (
            (b + c - a), (c + a - b), (a + b - c)
        )), **kwargs)


class SpiekerCenter(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        a, b, c = distance(B, C), distance(A, C), distance(A, B)
        super().__init__(trilinear_to_cartesian(A, B, C, (
            (b * c * (b + c)), (a * c * (a + c)), (b * a * (b + a))
        )), **kwargs)


class FeuerbachPoint(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        angle_a = Angle.from_three_points(B, A, C).get_value()
        angle_b = Angle.from_three_points(A, B, C).get_value()
        angle_c = Angle.from_three_points(B, C, A).get_value()
        super().__init__(trilinear_to_cartesian(A, B, C, (
            (1 - np.cos(angle_b - angle_c)), (1 - np.cos(angle_c - angle_a)), (1 - np.cos(angle_a - angle_b))
        )), **kwargs)


class FermatPoint(Dot):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        angle_a = Angle.from_three_points(B, A, C).get_value()
        angle_b = Angle.from_three_points(A, B, C).get_value()
        angle_c = Angle.from_three_points(B, C, A).get_value()
        if angle_a > 2 / 3 * PI:
            super().__init__(A)

        elif angle_b > 2 / 3 * PI:
            super().__init__(B)

        elif angle_c > 2 / 3 * PI:
            super().__init__(C)

        super().__init__(trilinear_to_cartesian(A, B, C, (
            csc(angle_a + PI / 3), csc(angle_b + PI / 3), csc(angle_c + PI / 3)
        )), **kwargs)



from manim.typing import Point3DLike
from manim.constants import *

from ..useful_in_development import *
import numpy as np


class Line(manim.Line):
    def __init__(self, start: Point3DLike = LEFT, end: Point3DLike = RIGHT, **kwargs):
        start, end = dot_to_array(start, end)
        super().__init__(start, end, **kwargs)

    def is_point_in_line(self, dot: Point3DLike):
        dot = dot_to_array(dot)[0]
        x1, y1 = self.get_start()[0], self.get_start()[1]
        x2, y2 = self.get_end()[0], self.get_end()[1]
        x3, y3 = dot[0], dot[1]
        return round(np.linalg.det(np.array([[x1, y1, 1], [x2, y2, 1], [x3, y3, 1]])), 3) == 0

    def set_length_about_point(self, length: float, dot: Point3DLike):
        dot = dot_to_array(dot)[0]
        A = dot + np.array([length, 0, 0])
        line = Line(dot, A).rotate(about_point=dot, angle=self.get_angle())
        self.put_start_and_end_on(line.get_start(), line.get_end())
        return self

    def get_distance(self, dot: Point3DLike):
        dot = dot_to_array(dot)[0]
        return Line(dot, self.get_projection(dot)).get_length()

    def equal(self, n: int = 1, buff: float = 0.1, length: float = 0.2, **kwargs):
        return manim.VGroup(*[Line(**kwargs).set_length(length).rotate(
            about_point=Line(**kwargs).set_length(length).get_center(), angle=np.pi / 2) for i in range(n)]).arrange(
            buff=buff, direction=RIGHT).move_to(self.get_center()).rotate(
            about_point=self.get_center(), angle=self.get_angle()).set_z_index(1)

    def paral(self, n: int = 1, buff: float = 0.08, length: float = 0.15, rotate: bool = False, **kwargs):
        elem = manim.VGroup(
            Line(**kwargs).set_length(length).rotate(about_point=Line().set_length(0.2).get_end(),
                                                     angle=PI / 4).shift(0.02 * UR),
            Line(**kwargs).set_length(length).rotate(about_point=Line().set_length(0.2).get_end(),
                                                     angle=-PI / 4).shift(0.02 * DR)
        ).rotate(int(rotate) * np.pi)
        return manim.VGroup(*[elem.copy() for i in range(n)]).arrange(buff=buff, direction=RIGHT).move_to(
            self.get_center()).rotate(about_point=self.get_center(), angle=self.get_angle()).set_z_index(1)

    def inversion(self, circle, **kwargs):
        if self.is_point_in_line(circle.get_center()):
            return Line(self.get_start(), self.get_end(), **kwargs)
        else:
            dot_1 = circle_symmetry(circle, self.point_from_proportion(0.3))
            dot_2 = circle_symmetry(circle, self.point_from_proportion(0.5))
            dot_3 = circle_symmetry(circle, self.point_from_proportion(0.7))
            return Circle().from_three_points(dot_1, dot_2, dot_3, **kwargs)


class Angle(manim.Angle):
    def __init__(self, *args, **kwargs):
        radius = kwargs.get("radius", None)
        quadrant = kwargs.get("quadrant", [1, 1])
        inter = manim.line_intersection(
            [args[0].get_start(), args[0].get_end()],
            [args[1].get_start(), args[1].get_end()],
        )
        line1 = args[0]
        line2 = args[1]
        if radius is None:
            if quadrant[0] == 1:
                dist_1 = np.linalg.norm(line1.get_end() - inter)
            else:
                dist_1 = np.linalg.norm(line1.get_start() - inter)
            if quadrant[1] == 1:
                dist_2 = np.linalg.norm(line2.get_end() - inter)
            else:
                dist_2 = np.linalg.norm(line2.get_start() - inter)
            if np.minimum(dist_1, dist_2) < 0.6:
                radius = (2 / 3) * np.minimum(dist_1, dist_2)
            else:
                radius = 0.4
            self.radius = radius
        else:
            self.radius = radius
        kwargs["radius"] = self.radius
        super().__init__(*args, z_index=-1, **kwargs)

    @staticmethod
    def from_three_points(A: Point3DLike, B: Point3DLike, C: Point3DLike, **kwargs):
        angle = Angle(Line(B, A), Line(B, C), **kwargs)
        try:
            if angle.get_angle_value() > PI:
                angle = Angle(Line(B, C), Line(B, A), **kwargs)
        except AttributeError:
            if Angle(Line(B, A), Line(B, C)).get_value() > PI:
                angle = Angle(Line(B, C), Line(B, A), **kwargs)
        return angle

    def get_label_center(self, delta: float = 0.25):
        A = self.lines[0].get_end()
        B = self.lines[0].get_start()
        C = self.lines[1].get_end()
        A, B, C = dot_to_array(A, B, C)

        try:
            radius = self.radius
        except AttributeError:
            inter = manim.line_intersection(
                [self.lines[0].get_start(), self.lines[0].get_end()],
                [self.lines[1].get_start(), self.lines[1].get_end()],
            )

            if self.quadrant[0] == 1:
                dist_1 = np.linalg.norm(self.lines[0].get_end() - inter)
            else:
                dist_1 = np.linalg.norm(self.lines[0].get_start() - inter)
            if self.quadrant[1] == 1:
                dist_2 = np.linalg.norm(self.lines[1].get_end() - inter)
            else:
                dist_2 = np.linalg.norm(self.lines[1].get_start() - inter)
            if np.minimum(dist_1, dist_2) < 0.6:
                radius = (2 / 3) * np.minimum(dist_1, dist_2)
            else:
                radius = 0.4

        center = Line(B, Line(A, C).point_from_proportion(
            ((Line(A, C).get_length() * Line(A, B).get_length()) /
             (Line(B, C).get_length() + Line(A, B).get_length())) /
            Line(A, C).get_length()
        )).set_length_about_point(delta + radius, B).get_end()

        return center


class Circle(manim.Circle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def from_three_points(p1: Point3DLike, p2: Point3DLike, p3: Point3DLike, **kwargs):
        p1, p2, p3 = dot_to_array(p1, p2, p3)
        center = manim.line_intersection(
            manim.perpendicular_bisector([p1, p2]),
            manim.perpendicular_bisector([p2, p3]),
        )
        radius = np.linalg.norm(p1 - center)
        return Circle(radius=radius, **kwargs).shift(center)

    def pow(self, dot: Point3DLike):
        dot = dot_to_array(dot)[0]
        return round((pow(self.get_center()[0] - dot[0], 2) + pow(self.get_center()[1] - dot[1], 2)) -
                     self.radius ** 2, 4)

    def is_point_in_circle(self, dot: Point3DLike):
        dot = dot_to_array(dot)[0]
        return self.pow(dot) == 0

    def get_diametrically_opposite_point(self, dot: Point3DLike):
        dot = dot_to_array(dot)[0]
        return self.get_center() + self.get_center() - dot


class Dot(manim.Dot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, z_index=1, **kwargs)



from __future__ import annotations

import numpy as np
from .default import Line, Circle, Dot

__all__ = [
    "intersection_lines",
    "intersection_circles",
    "intersection_line_and_circle"
]


def intersection_lines(line_1: Line, line_2: Line) -> np.ndarray:
    """
    Find the intersection point of two lines.

    :param line_1:
    :param line_2:
    :return:
    """
    x1, y1 = line_1.get_start()[0], line_1.get_start()[1],
    x2, y2 = line_1.get_end()[0], line_1.get_end()[1],
    x3, y3 = line_2.get_start()[0], line_2.get_start()[1],
    x4, y4 = line_2.get_end()[0], line_2.get_end()[1],
    try:
        x0 = ((x1 * y2 - y1 * x2) * (x3 - x4) - (x1 - x2) * (x3 * y4 - y3 * x4)) / (
            (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4))
        y0 = ((x1 * y2 - y1 * x2) * (y3 - y4) - (y1 - y2) * (x3 * y4 - y3 * x4)) / (
            (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4))
    except ZeroDivisionError:
        raise Exception("Lines are parallel")
    return np.array([x0, y0, 0])


def intersection_circles(circle_1: Circle, circle_2: Circle) -> tuple[np.ndarray, np.ndarray]:
    """
    Find the intersection points of two circles.

    :param circle_1:
    :param circle_2:
    :return:
    """
    o1 = circle_1.get_center()
    o2 = circle_2.get_center()
    r1 = circle_1.radius
    r2 = circle_2.radius
    l = Line(o1, o2).get_length()

    if l > r1 + r2 + 0.02:
        raise Exception('Circles do not intersect')

    x1 = Dot(Line(o1, o2).set_length_about_point(r1, o1).get_end())
    x2 = x1.copy()

    alpha = np.arccos(round((r2 ** 2 - r1 ** 2 - l ** 2) / (-2 * r1 * l), 4))
    x1.rotate(about_point=o1, angle=alpha)
    x2.rotate(about_point=o1, angle=-alpha)

    return x1.get_center(), x2.get_center()


def intersection_line_and_circle(line: Line, circle: Circle) -> tuple[np.ndarray, np.ndarray]:
    """
    Find the intersection points of a line and a circle.

    :param line:
    :param circle:
    :return:
    """
    o, r = circle.get_center(), circle.radius
    h = Line(line.get_projection(o), o).get_length()
    alpha = np.arccos(round(h / r, 4))
    x1 = Dot(Line(o, line.get_projection(o)).set_length_about_point(r, o).get_end())
    x2 = x1.copy()
    x1.rotate(about_point=o, angle=alpha)
    x2.rotate(about_point=o, angle=-alpha)
    return x1.get_center(), x2.get_center()



import typing

from manim import AnnularSector, VGroup, Tex
from manim.constants import *
from manim.utils.color import *

import numpy as np
import typing

PIE_CHART_COLORS = [
    RED,
    GREEN,
    BLUE,
    MAROON,
    YELLOW,
    TEAL
]


class PieChartSector(AnnularSector):
    """
    A sector of a pie chart.

    Parameters
    ----------
    value : float
        The value of the sector.
    percent : float
        The percentage of the sector.
    label : str
        The label of the sector.
    """
    def __init__(self, value, percent, label, *args, **kwargs):
        self.value = value
        self.percent = percent
        self.label = label
        angle = percent / 100 * 2 * PI
        super().__init__(angle=angle + 0.0075, *args, **kwargs)


class PieChart(VGroup):
    """
    A pie chart.

    Parameters
    ----------
    data : Union[dict, tuple]
        The data of the pie chart. If it is a dictionary, the keys are the labels and the values are the values.
        If it is a tuple, the first element is the labels and the second element is the values.
    labels : Optional[List[str]]
        The labels of the pie chart.
    label_buff : float, optional
        The buffer between the label and the sector. Default to 0.85.
    colors : List[Union[str, Tuple[float, float, float]]], optional
        The colors of the sectors. Default to PIE_CHART_COLORS.
    inner_radius : float, optional
        The inner radius of the sectors. Default to 1.
    outer_radius : float, optional
        The outer radius of the sectors. Default to 2.

    Examples
    --------
    .. manimextra:: PieChartExample
        :save_last_frame:

        class PieChartExample(Scene):
            def construct(self):
                pie_chart = PieChart(
                    data={"First": 10, "Second": 20, "Third": 30},
                    label_buff=0.7,
                    inner_radius=1,
                    outer_radius=2
                )
                self.add(pie_chart)
    """
    def __init__(self,
                 data,
                 labels=None,
                 label_buff: float = 0.85,
                 colors=PIE_CHART_COLORS,
                 inner_radius=1,
                 outer_radius=2,
                 ):
        if isinstance(data, dict):
            labels = list(data.keys())
            values = list(data.values())
        elif isinstance(data, tuple):
            labels, values = data
        else:
            labels = labels if labels else [i for i in range(len(data))]
            values = data

        n = len(values)
        sum_values = sum(values)
        sum_angle = 0

        self.values = values
        self.labels = VGroup()
        self.sectors = VGroup()

        if n > len(colors):
            raise ValueError(f"Too many values for pie chart. Maximum is {len(PIE_CHART_COLORS)}")

        for i in range(n):
            val = values[i]
            angle = val / sum_values * 2 * PI

            sector = PieChartSector(
                value=val,
                percent=100 * val / sum_values,
                label=labels[i],
                color=colors[i],
                inner_radius=inner_radius,
                outer_radius=outer_radius
            ).rotate(about_point=ORIGIN, angle=sum_angle)

            label = Tex(labels[i]).move_to(RIGHT * sector.outer_radius).rotate(
                about_point=ORIGIN, angle=(sum_angle + angle / 2)).rotate(-(sum_angle + angle / 2)).shift(
                label_buff * np.array(
                    [np.cos(sum_angle + angle / 2), np.sin(sum_angle + angle / 2), 0]))

            sum_angle += angle

            self.sectors.add(sector)
            self.labels.add(label)

        super().__init__(self.sectors, self.labels)

    def get_labels(self):
        return self.labels

    def get_sectors(self):
        return self.sectors

    def get_sector(self, value=None, label=None):
        """
        Get a sector by value or label.

        :param value:
        :param label:
        :return:
        """
        for sector in self.sectors:
            if value:
                if sector.value == value:
                    return sector
            if label:
                if sector.label == label:
                    return sector
        raise ValueError("No sector with that value or label found")



from __future__ import annotations

from manim.constants import *

from .default import *
from .intersection import intersection_circles
from ..useful_in_development import *
from .triangle_centers import *


__all__ = [
    "Tangent",
    "Incircle",
    "Circumcircle",
    "NinePointCircle",
    "CarnotCircle"

]


class Tangent(Line):
    """
    A class to represent a tangent to a circle.

    Parameters
    ----------
    circle : Circle
        The circle to which the tangent is drawn.
    dot : Dot
        The point from which the tangent is drawn. It can be on the circle or outside the circle.
    is_other_tangent : bool, optional
        If True, the tangent is drawn from the other intersection point of the circle and the line, by default False.
    """
    def __init__(self, circle: Circle, dot: Dot, is_other_tangent=False, **kwargs):
        dot = dot_to_array(dot)[0]
        if circle.pow(dot) < 0:
            raise Exception("Dot is inside the circle")

        self.circle = circle
        self.dot = dot
        self.tangent_point = Dot(intersection_circles(
            circle, Circle(arc_center=dot, radius=np.sqrt(
                pow(distance(circle.get_center(), dot), 2) - pow(circle.get_radius(), 2)
            ))
        )[is_other_tangent])
        self.is_other_tangent = is_other_tangent
        super().__init__(dot, self.tangent_point, **kwargs)

    def get_other_tangent(self):
        return Tangent(self.circle, self.dot, is_other_tangent=(not self.is_other_tangent))


class Incircle(Circle):
    """
    A class to represent the incircle of a triangle.

    Parameters
    ----------
    A : np.ndarray
        The first vertex of the triangle.
    B : np.ndarray
        The second vertex of the triangle.
    C : np.ndarray
        The third vertex of the triangle.
    """
    def __init__(self, A, B, C, *args, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        I = Incenter(A, B, C).get_center()
        a = distance(B, C)
        b = distance(A, C)
        c = distance(A, B)
        p = (a + b + c) / 2
        r = np.sqrt((p - a) * (p - b) * (p - c) / p)
        super().__init__(radius=r, arc_center=I, *args, **kwargs)


class Circumcircle(Circle):
    """
    A class to represent the circumcircle of a triangle.

    Parameters
    ----------
    A : np.ndarray
        The first vertex of the triangle.
    B : np.ndarray
        The second vertex of the triangle.
    C : np.ndarray
        The third vertex of the triangle.
    """
    def __init__(self, A, B, C, *args, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        O = Circumcenter(A, B, C).get_center()
        r = distance(O, A)
        super().__init__(radius=r, arc_center=O, *args, **kwargs)


class NinePointCircle(Circle):
    """
    A class to represent the nine-point circle of a triangle.

    Parameters
    ----------
    A : np.ndarray
        The first vertex of the triangle.
    B : np.ndarray
        The second vertex of the triangle.
    C : np.ndarray
        The third vertex of the triangle.
    """
    def __init__(self, A, B, C, *args, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        r = distance(Circumcenter(A, B, C).get_center(), A) / 2
        print(r)
        O = NinePointCenter(A, B, C).get_center()
        super().__init__(radius=r, arc_center=O, *args, **kwargs)


class CarnotCircle(Circle):
    """
    A class to represent the Carnot circle of a triangle relative to side AC.

    Parameters
    ----------
    A : np.ndarray
        The first vertex of the triangle.
    B : np.ndarray
        The second vertex of the triangle.
    C : np.ndarray
        The third vertex of the triangle.
    """
    def __init__(self, A, B, C, *args, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        circle = Circumcircle(A, B, C)
        circle.rotate(PI, about_point=Line(A, C).get_projection(circle.get_center()))
        super().__init__(radius=circle.get_radius(), arc_center=circle.get_center(), *args, **kwargs)



from manim import VGroup, BraceBetweenPoints
from manim.constants import *

from .bracket import BracketBetweenPoints

__all__ = [
    "SystemOfEquations"
]


class SystemOfEquations(VGroup):
    """
    A group of equations with a brace on the left side.

    Parameters
    ----------
    *equations : :class:`manim.mobject.Mobject`
        The equations.
    is_bracket : bool, optional
        Whether to use bracket instead of brace. Default to False.
    buff_between_equations : float, optional
        The buffer between equations. Default to 0.2.
    buff_between_brace_and_equations : float, optional
        The buffer between brace and equations. Default to 0.05.

    Examples
    --------
    .. manimextra:: SystemOfEquationsExample
        :save_last_frame:

        class SystemOfEquationsExample(Scene):
            def construct(self):
                system = SystemOfEquations(
                    MathTex("x^2", "+", "y^2", "=", "9"),
                    MathTex("y", "=", "x", "+", "1")
                )
                self.add(system)
    """
    def __init__(
            self,
            *equations,
            is_bracket=False,
            buff_between_equations=0.2,
            buff_between_brace_and_equations=0.05,
    ):
        self.equations = VGroup(*equations).arrange(DOWN, buff=buff_between_equations)
        self.is_bracket = is_bracket
        self.buff_between_equations = buff_between_equations
        self.buff_between_brace_and_equations = buff_between_brace_and_equations

        max_width = max([equ.get_width() for equ in self.equations])
        for equ in self.equations:
            delta = (max_width - equ.get_width()) / 2
            equ.shift(LEFT * delta)

        if not is_bracket:
            self.brace = BraceBetweenPoints(
                self.equations[0].get_corner(UP + LEFT) + 0.1 * UP,
                self.equations[-1].get_corner(DOWN + LEFT) + 0.1 * DOWN,
            )
        else:
            self.brace = BracketBetweenPoints(
                self.equations[0].get_corner(UP + LEFT) + 0.1 * UP,
                self.equations[-1].get_corner(DOWN + LEFT) + 0.1 * DOWN,
            )

        self.brace.next_to(self.equations, LEFT, buff=buff_between_brace_and_equations)
        super().__init__(self.equations, self.brace)

    def swap(self, i, j):
        """
        Swap the i-th and j-th equations.
        :param i:
        :param j:
        :return:
        """
        self.equations[i], self.equations[j] = self.equations[j], self.equations[i]

        self.equations.arrange(DOWN, buff=self.buff_between_equations)
        max_width = max([equ.get_width() for equ in self.equations])
        for equ in self.equations:
            delta = (max_width - equ.get_width()) / 2
            equ.shift(LEFT * delta)

        self.brace.next_to(self.equations, LEFT, buff=self.buff_between_brace_and_equations)
        return self



import math

from manim import (
    MathTex,
    VGroup,
    Arc,
    FadeOut,
    AnimationGroup,
    override_animate,
    FadeIn
)

from manim.utils.color.manim_colors import *

from ..geometry.default import *

__all__ = [
    "UnitCircleLabel",
    "UnitCircle"
]


class UnitCircleLabel(MathTex):
    """
    A label for the unit circle.

    Parameters
    ----------
    direction : np.ndarray
        The direction of the label.
    repeats : int
        The number of repeats of the unit circle.
    fraction : bool
        Whether to use fraction. Default to True.
    """
    def __init__(self, direction=RIGHT, repeats=0, fraction=True, **kwargs):
        if (direction == RIGHT).all():
            self.other = False
            if repeats == 0:
                super().__init__(r'0', **kwargs)
            else:
                super().__init__(str(2 * repeats) + r'\pi', **kwargs)
        elif (direction == LEFT).all():
            if repeats == 0:
                super().__init__(r'\pi', **kwargs)
            elif repeats == -1:
                super().__init__(r'-\pi', **kwargs)
            else:
                super().__init__(str(2 * repeats + 1) + r'\pi', **kwargs)

        if fraction:
            if (direction == UP).all():
                if repeats > 0:
                    up_label_text = r'\dfrac{' + str(4 * repeats + 1) + r'\pi}{2}'
                elif repeats == 0:
                    up_label_text = r'\dfrac{\pi}{2}'
                else:
                    up_label_text = r'-\dfrac{' + str(-(4 * repeats + 1)) + r'\pi}{2}'
                super().__init__(up_label_text, **kwargs)
            if (direction == DOWN).all():
                if repeats >= 0:
                    down_label_text = r'\dfrac{' + str(4 * repeats + 3) + r'\pi}{2}'
                elif repeats == -1:
                    down_label_text = r'-\dfrac{\pi}{2}'
                else:
                    down_label_text = r'-\dfrac{' + str(-(4 * repeats + 3)) + r'\pi}{2}'
                super().__init__(down_label_text, **kwargs)

        else:
            if (direction == UP).all():
                super().__init__(str(2 * repeats + 0.5) + r'\pi', **kwargs)
            elif (direction == DOWN).all():
                super().__init__(str(2 * repeats + 1.5) + r'\pi', **kwargs)


class FadeInAndAdd(AnimationGroup):
    def __init__(self, unit_circle, *objects, **kwargs):
        self.unit_circle = unit_circle
        self.unit_circle.normalize_labels()

        animations = []
        for obj in objects:
            animations.append(FadeIn(obj))

        super().__init__(*animations, **kwargs)


class FadeOutAndRemove(AnimationGroup):
    def __init__(self, unit_circle, *objects, **kwargs):
        self.unit_circle = unit_circle
        self.unit_circle.normalize_labels()

        animations = []
        for obj in objects:
            animations.append(FadeOut(obj))

        super().__init__(*animations, **kwargs)


class UnitCircle(VGroup):
    """
    A unit circle.

    Parameters
    ----------
    point : float
        The point on the unit circle. Using this point we determine how many cycles have passed.
    radius : float, optional
        The radius of the circle. Default to 1.5.
    color : :class:`manim.utils.color.Color`, optional
        The color of the circle. Default to BLUE.
    label_buff : float, optional
        The buffer between the circle and the labels. Default to 0.2.
    font_size : int, optional
        The font size of the labels. Default to 32.
    fractions : bool, optional
        Whether to use fraction in the labels. Default to True.

    Examples
    --------
    .. manimextra:: UnitCircleExample
        :save_last_frame:

        class UnitCircleExample(Scene):
            def construct(self):
                circle = UnitCircle()
                circle.add_point(PI / 2, "\dfrac{\pi}{2}")
                circle.add_point(PI, "\pi")
                self.add(circle)
    """
    def __init__(self, point=0, radius=1.5, color=BLUE, label_buff=0.2, font_size=32, fractions=True):
        self.circle = Circle(radius=radius, color=color)
        self.circle.set_z_index(-5)

        self.labels = VGroup()
        self.dots = VGroup()

        self.other_right_basic_label = False

        repeats = math.floor(point / (2 * np.pi))
        self.repeats = repeats

        self.label_buff = label_buff
        self.font_size = font_size

        right = Dot(self.circle.point_at_angle(0))
        up = Dot(self.circle.point_at_angle(np.pi / 2))
        left = Dot(self.circle.point_at_angle(np.pi))
        down = Dot(self.circle.point_at_angle(3 * np.pi / 2))

        self.right = right
        self.up = up
        self.left = left
        self.down = down

        self.right_label = UnitCircleLabel(direction=RIGHT, repeats=repeats, fraction=fractions,
                                           font_size=font_size).next_to(right, RIGHT, buff=label_buff)
        self.up_label = UnitCircleLabel(direction=UP, repeats=repeats, fraction=fractions, font_size=font_size).next_to(
            up, UP, buff=label_buff)
        self.left_label = UnitCircleLabel(direction=LEFT, repeats=repeats, fraction=fractions,
                                          font_size=font_size).next_to(left, LEFT, buff=label_buff)
        self.down_label = UnitCircleLabel(direction=DOWN, repeats=repeats, fraction=fractions,
                                          font_size=font_size).next_to(down, DOWN, buff=label_buff)

        self.horizontal = Line(self.left, self.right)
        self.vertical = Line(self.down, self.up)

        super().__init__(self.circle, self.labels, self.dots)

    def change_right_point(self):
        if self.other_right_basic_label:
            self.right_label = UnitCircleLabel(direction=RIGHT, repeats=self.repeats, font_size=self.font_size).next_to(
                self.right, RIGHT, buff=self.label_buff)
        else:
            self.right_label = UnitCircleLabel(direction=RIGHT, repeats=self.repeats + 1,
                                               font_size=self.font_size).next_to(
                self.right, RIGHT, buff=self.label_buff)

        self.other_right_basic_label = not self.other_right_basic_label
        return self

    def get_center(self):
        """
        Get the center of the circle.
        Override because the center of VGroup is different from the center of the circle.
        """
        return self.circle.get_center()

    def add_point(self, point, label):
        """
        Add a point to the unit circle.

        Parameters
        ----------
        point
        label

        Returns
        -------

        """
        dot = Dot(self.circle.point_at_angle(point))
        if isinstance(label, str):
            print(Line(self.get_center(), dot.get_center()).get_unit_vector())
            label = MathTex(label,
                            font_size=self.font_size
                            ).next_to(dot,
                                      np.round(Line(self.get_center(), dot.get_center()).get_unit_vector(), 1),
                                      buff=self.label_buff
                                      )
        self.dots.add(dot)
        self.labels.add(label)
        return VGroup(dot, label)

    @override_animate(add_point)
    def _add_point_animation(self, point, label, anim_args=None):
        if anim_args is None:
            anim_args = {}
        dot, label = self.add_point(point, label)
        anim = FadeIn(VGroup(dot, label), **anim_args)
        return anim

    def normalize_labels(self):
        VGroup(self.right, self.up, self.left, self.down, self.right_label, self.up_label, self.left_label,
               self.down_label).move_to(self.circle.get_center())

    def show_labels(self):
        self.normalize_labels()
        self.add(self.right, self.up, self.left, self.down, self.right_label, self.up_label, self.left_label,
                 self.down_label)
        return self

    def hide_labels(self):
        self.remove(self.right, self.up, self.left, self.down, self.right_label, self.up_label, self.left_label,
                    self.down_label)
        return self

    @override_animate(show_labels)
    def _show_labels_animation(self, anim_args=None):
        self.normalize_labels()
        if anim_args is None:
            anim_args = {}
        return FadeInAndAdd(self, self.right, self.up, self.left, self.down,
                            self.right_label, self.up_label, self.left_label, self.down_label, **anim_args)

    @override_animate(hide_labels)
    def _hide_labels_animation(self, anim_args=None):
        if anim_args is None:
            anim_args = {}
        return FadeOutAndRemove(self, self.right, self.up, self.left, self.down,
                                self.right_label, self.up_label, self.left_label, self.down_label, **anim_args)

    def show_horizontal(self):
        self.add(self.horizontal)
        return self

    def hide_horizontal(self):
        self.remove(self.horizontal)
        return self

    @override_animate(show_horizontal)
    def _show_horizontal_animation(self, anim_args=None):
        if anim_args is None:
            anim_args = {}
        return FadeInAndAdd(self, self.horizontal, **anim_args)

    @override_animate(hide_horizontal)
    def _hide_horizontal_animation(self, anim_args=None):
        if anim_args is None:
            anim_args = {}
        return FadeOutAndRemove(self, self.horizontal, **anim_args)

    def show_vertical(self):
        self.add(self.vertical)
        return self

    def hide_vertical(self):
        self.remove(self.vertical)
        return self

    @override_animate(show_vertical)
    def _show_vertical_animation(self, anim_args=None):
        if anim_args is None:
            anim_args = {}
        return FadeInAndAdd(self, self.vertical, **anim_args)

    @override_animate(hide_vertical)
    def _hide_vertical_animation(self, anim_args=None):
        if anim_args is None:
            anim_args = {}
        return FadeOutAndRemove(self, self.vertical, **anim_args)

    def show_axes(self):
        self.add(self.horizontal, self.vertical)
        return self

    def hide_axes(self):
        self.remove(self.horizontal, self.vertical)
        return self

    @override_animate(show_axes)
    def _show_axes_animation(self, anim_args=None):
        if anim_args is None:
            anim_args = {}
        return FadeInAndAdd(self, self.horizontal, self.vertical, **anim_args)

    @override_animate(hide_axes)
    def _hide_axes_animation(self, anim_args=None):
        if anim_args is None:
            anim_args = {}
        return FadeOutAndRemove(self, self.horizontal, self.vertical, **anim_args)

    def get_abscissa_point(self, point):
        alpha = (1 + point) / 2
        return self.horizontal.point_from_proportion(alpha)

    def get_ordinate_point(self, point):
        alpha = (1 + point) / 2
        return self.vertical.point_from_proportion(alpha)

    def get_arc(self, start, end, **kwargs):
        if start > end:
            start, end = end, start

        angle = end - start
        return Arc(arc_center=self.circle.get_center(), radius=self.circle.get_radius(),
                   start_angle=start, angle=angle,
                   **kwargs)

    def point_at_angle(self, angle):
        return self.circle.point_at_angle(angle)



from __future__ import annotations

from manim.typing import Point3DLike

from ..useful_in_development import *

from .default import Line, Angle
from manim import VGroup
from manim.constants import PI

__all__ = [
    "Cevian",
    "Bisector",
    "Median",
    "Perpendicular",
    "Altitude",
    "PerpendicularBisector",
    "Symmedian",
    "EuclidLine",
]


class Cevian(Line):
    """
    A class to represent a cevian in a triangle.

    Parameters
    ----------
    A : Point3DLike
        The first vertex of the triangle.
    B : Point3DLike
        The second vertex of the triangle. Cevian passes through this point.
    C : Point3DLike
        The third vertex of the triangle.
    alpha : float, optional
        The proportion of the cevian, by default 0.35. Let D be the foot of the cevian, then the proportion is the AD:AC ratio.
    """

    def __init__(self, A: Point3DLike, B: Point3DLike, C: Point3DLike, alpha=0.35, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        D = Line(A, C).point_from_proportion(alpha)

        self.foot = D
        self.general_vertex = B
        self.extra_vertex_1 = A
        self.extra_vertex_2 = C
        self.proportion = alpha

        super().__init__(B, D, **kwargs)

    def get_foot(self) -> Point3DLike:
        return self.foot

    def get_vertex(self) -> Point3DLike:
        return self.general_vertex

    def get_proportion(self) -> float:
        return self.proportion


class Bisector(Cevian):
    """
    A class to represent a bisector in a triangle.

    Parameters
    ----------
    A : Point3DLike
        The first vertex of the triangle.
    B : Point3DLike
        The second vertex of the triangle.
    C : Point3DLike
        The third vertex of the triangle.

    Examples
    ----------
    .. manimextra:: BisectorExample
        :save_last_frame:

        class BisectorExample(Scene):
            def construct(self):
                A = Dot(1.5 * DOWN + 4 * LEFT).set_z_index(1)
                B = Dot(1.5 * UP + 1.5 * LEFT).set_z_index(1)
                C = Dot(1.5 * DOWN + 4 * RIGHT)

                a = Line(B, C, color=BLUE)
                b = Line(A, C, color=BLUE)
                c = Line(A, B, color=BLUE)

                bisector = Bisector(A, B, C, color=YELLOW)
                angles = bisector.get_bisected_angles()
                self.add(A, B, C, a, b, c, bisector, angles)
    """

    def __init__(self, A: Point3DLike, B: Point3DLike, C: Point3DLike, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        alpha = ((Line(A, C).get_length() * Line(A, B).get_length()) /
                 (Line(B, C).get_length() + Line(A, B).get_length())) / Line(A, C).get_length()
        super().__init__(A, B, C, alpha, **kwargs)

    def get_bisected_angles(self, radius_alpha=1.1, **kwargs) -> VGroup(Angle, Angle):
        first_angle = Angle.from_three_points(self.extra_vertex_1, self.general_vertex, self.foot, **kwargs)
        kwargs['radius'] = first_angle.radius * radius_alpha
        second_angle = Angle.from_three_points(self.extra_vertex_2, self.general_vertex, self.foot, **kwargs)
        return VGroup(first_angle, second_angle)


class Median(Cevian):
    """
    A class to represent a median in a triangle.

    Parameters
    ----------
    A : Point3DLike
        The first vertex of the triangle.
    B : Point3DLike
        The second vertex of the triangle.
    C : Point3DLike
        The third vertex of the triangle.

    Examples
    ----------
    .. manimextra:: MedianExample
        :save_last_frame:

        class MedianExample(Scene):
            def construct(self):
                A = Dot(DOWN + 4 * LEFT).set_z_index(1)
                B = Dot(1.5 * UP + 1.5 * LEFT).set_z_index(1)
                C = Dot(DOWN + 4 * RIGHT)

                a = Line(B, C, color=BLUE)
                b = Line(A, C, color=BLUE)
                c = Line(A, B, color=BLUE)

                median = Median(A, B, C, color=YELLOW)
                self.add(A, B, C, a, b, c, median)
    """

    def __init__(self, A: Point3DLike, B: Point3DLike, C: Point3DLike, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        super().__init__(A, B, C, 0.5, **kwargs)

    def get_equal_segments(self, **kwargs) -> VGroup(Line, Line):
        return VGroup(
            Line(self.extra_vertex_1, self.foot).equal(**kwargs),
            Line(self.extra_vertex_2, self.foot).equal(**kwargs),
        )


class Symmedian(Cevian):
    """
    A class to represent a symmedian in a triangle.

    Parameters
    ----------
    A : Point3DLike
        The first vertex of the triangle.
    B : Point3DLike
        The second vertex of the triangle.
    C : Point3DLike
        The third vertex of the triangle

    """

    def __init__(self, A: Point3DLike, B: Point3DLike, C: Point3DLike, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        a, b, c = Line(B, C).get_length(), Line(A, C).get_length(), Line(A, B).get_length()
        alpha = c ** 2 / (a ** 2 + c ** 2)
        super().__init__(A, B, C, alpha=alpha, **kwargs)


class Perpendicular(Line):
    """
    A class to represent a perpendicular line to another line.

    Parameters
    ----------
    line : Line
        The line to which the perpendicular line is drawn.
    dot : Point3DLike
        The point through which the perpendicular line passes.
    length : float, optional
        The length of the perpendicular line, by default 1.0.
    rotate : bool, optional
        Use if you want the perpendicular to be in the other direction, by default False.
    """

    def __init__(self, line: Line, dot, length=1.0, rotate=False, **kwargs):
        X = dot_to_array(dot)[0]
        A, B = line.get_start(), line.get_end()
        if Line(A, B).is_point_in_line(X):
            perpendicular = Line(A, X).rotate(
                about_point=X, angle=PI / 2 + int(rotate) * PI).set_length_about_point(length, X)
        else:
            perpendicular = Line(X, Line(A, B).get_projection(X))

        self.vertex = X
        self.foot = Line(A, B).get_projection(X)
        self.line = line

        super().__init__(perpendicular.get_start(), perpendicular.get_end(), **kwargs)

    def get_line(self) -> Line:
        return self.linez

    def get_vertex(self) -> Point3DLike:
        return self.vertex

    def get_foot(self) -> Point3DLike:
        return self.foot

    def get_angles(self, **kwargs):
        return VGroup(
            Angle.from_three_points(self.vertex, self.dot, self.line.get_start(), elbow=True, **kwargs),
            Angle.from_three_points(self.vertex, self.dot, self.line.get_end(), elbow=True, **kwargs)
        )


class PerpendicularBisector(Line):
    """
    A class to represent a perpendicular bisector to a line.

    Parameters
    ----------
    line : Line
        The line to which the perpendicular bisector is drawn.
    length : float, optional
        The length of the perpendicular bisector, by default 1.0.
    """

    def __init__(self, line: Line, length=1, **kwargs):
        line_1 = Perpendicular(line=line, dot=line.get_center(), length=length / 2, rotate=False)
        line_2 = Perpendicular(line=line, dot=line.get_center(), length=length / 2, rotate=True)

        self.dot = line.get_center()
        self.line = line

        super().__init__(line_1.get_end(), line_2.get_end(), **kwargs)

    def get_angles(self, **kwargs):
        """
        Get all the angles formed by the perpendicular bisector.

        :param kwargs:
        :return:
        """
        return VGroup(
            Angle.from_three_points(self.line.get_start(), self.dot, self.get_start(), **kwargs),
            Angle.from_three_points(self.line.get_start(), self.dot, self.get_end(), **kwargs),
            Angle.from_three_points(self.line.get_end(), self.dot, self.get_start(), **kwargs),
            Angle.from_three_points(self.line.get_end(), self.dot, self.get_end(), **kwargs),
        )


class Altitude(Perpendicular):
    """
    A class to represent an altitude in a triangle. Important: it is not a cevian.

    Parameters
    ----------
    A : Point3DLike
        The first vertex of the triangle.
    B : Point3DLike
        The second vertex of the triangle.
    C : Point3DLike
        The third vertex of the triangle.
    """

    def __init__(self, A: Point3DLike, B: Point3DLike, C: Point3DLike, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        line = Line(A, C)
        super().__init__(line, B, **kwargs)


class EuclidLine(Line):
    """
    A class to represent a EuclidLine for a triangle.This line is parallel to AC and we pass through B

    Parameters
    ----------
    A : Point3DLike
        The first vertex of the triangle.
    B : Point3DLike
        The second vertex of the triangle.
    C : Point3DLike
        The third vertex of the triangle.
    """

    def __init__(self, A: Point3DLike, B: Point3DLike, C: Point3DLike, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        line = Line(A, C).move_to(B)
        super().__init__(line.get_start(), line.get_end(), **kwargs)



from .geometry import *  # noqa: F401, F403
from .diagrams import *  # noqa: F401, F403
from .animations import *  # noqa: F401, F403
from .algebra import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403



from __future__ import annotations

import manim
from manim.constants import *
from manim.utils.deprecation import deprecated

__all__ = [
    "Fancy_label"
]


@deprecated(since="v1.11.4", message="Please, dont use this cringe.")
def Fancy_label(text: manim.Tex, mode='normal', buff=manim.MED_LARGE_BUFF):
    """
    A function that writes text at the very top and adapts to its length

    Examples:
        .. manimextra:: FancyLabelExample

            class FancyLabelExample(manim.Scene):
                def construct(self):
                    label = Tex("Hello, World!")
                    self.play(Fancy_label(label))
                    self.wait()
    """

    mode = mode.lower()
    modes = ['vlow', 'low', 'normal', 'fast', 'vfast']

    if mode not in modes:
        mode = 'normal'

    symbols = len(text.get_tex_string())
    time = (symbols / (9 + modes.index(mode))) + 0.25

    return manim.AnimationGroup(manim.Write(text.to_edge(UP, buff=buff).set_z_index(999)), run_time=time)
